@@ -8,15 +8,68 @@ import shutil
 import subprocess
 import urllib.parse
 import urllib.request
+import sys
 import tarfile
 import tempfile
 import zipfile
 
 import colorama
+import jsonschema
 import yaml
 
 verbosity = False
 
+global_yaml_loader = yaml.SafeLoader
+global_bootstrap_validator = None
+native_yaml_available = False
+
+try:
+	global_yaml_loader = yaml.CSafeLoader
+	native_yaml_available = True
+except AttributeError:
+	pass
+
+def load_bootstrap_yaml(path):
+	global global_bootstrap_validator
+	if not global_bootstrap_validator:
+		schema_path = os.path.join(os.path.dirname(__file__), 'schema.yml')
+		with open(schema_path, 'r') as f:
+			schema_yml = yaml.load(f, Loader=global_yaml_loader)
+		global_bootstrap_validator = jsonschema.Draft7Validator(schema_yml)
+
+	with open(path, 'r') as f:
+		yml = yaml.load(f, Loader=global_yaml_loader)
+
+	any_errors = False
+	n = 0
+	for e in global_bootstrap_validator.iter_errors(yml):
+		if n == 0:
+			print("{}xbstrap{}: {}Failed to validate boostrap.yml{}".format(
+					colorama.Style.BRIGHT, colorama.Style.NORMAL,
+					colorama.Fore.RED, colorama.Style.RESET_ALL),
+					file=sys.stderr)
+		print("         {}* Error in file: {}, YAML element: {}\n"
+				"           {}{}".format(colorama.Fore.RED,
+				path, '/'.join(str(elem) for elem in e.absolute_path), e.message,
+				colorama.Style.RESET_ALL),
+				file=sys.stderr)
+		any_errors = True
+		n += 1
+		if n >= 10:
+			print("{}xbstrap{}: {}Reporting only the first 10 errors{}".format(
+					colorama.Style.BRIGHT, colorama.Style.NORMAL,
+					colorama.Fore.RED, colorama.Style.RESET_ALL),
+					file=sys.stderr)
+			break
+
+	if any_errors:
+		print("{}xbstrap{}: {}Validation issues will become hard errors in the future{}".format(
+				colorama.Style.BRIGHT, colorama.Style.NORMAL, colorama.Fore.YELLOW,
+				colorama.Style.RESET_ALL),
+				file=sys.stderr)
+
+	return yml
+
 def touch(path):
 	with open(path, 'w') as f:
 		pass
@@ -75,7 +128,12 @@ def installtree(src_root, dest_root):
 		src_path = os.path.join(src_root, name)
 		dest_path = os.path.join(dest_root, name)
 
-		if os.path.isdir(src_path):
+		# We do islink before isdir, as isdir resolves symlinks
+		if os.path.islink(src_path):
+			try_unlink(dest_path)
+			# Do not preserve attributes
+			os.symlink(os.readlink(src_path), dest_path)
+		elif os.path.isdir(src_path):
 			if not os.access(dest_path, os.F_OK):
 				# We only copy attributes when the directory is first created.
 				os.mkdir(dest_path)
@@ -84,11 +142,7 @@ def installtree(src_root, dest_root):
 			installtree(src_path, dest_path)
 		else:
 			try_unlink(dest_path)
-			if os.path.islink(src_path):
-				# Do not preserve attributes
-				os.symlink(os.readlink(src_path), dest_path)
-			else:
-				shutil.copy2(src_path, dest_path)
+			shutil.copy2(src_path, dest_path)
 
 class ResetMode(Enum):
 	NONE = 0
@@ -112,18 +166,33 @@ class ItemState:
 		self.timestamp = timestamp
 
 class Config:
-	def __init__(self, root_yml, site_yml):
-		self._root_yml = root_yml
-		self._site_yml = site_yml
+	def __init__(self, path):
+		self._config_path = path
+		self._root_yml = None
+		self._site_yml = dict()
 		self._sources = dict()
 		self._tool_pkgs = dict()
 		self._tool_stages = dict()
 		self._target_pkgs = dict()
 		self._tasks = dict()
 
-		self._parse_yml(self._root_yml)
+		self._bootstrap_path = os.path.join(path,
+				os.path.dirname(os.readlink(os.path.join(path, 'bootstrap.link'))))
 
-	def _parse_yml(self, current_yml, filter_sources = None, filter_tools = None, filter_pkgs = None, filter_tasks = None):
+		root_path = os.path.join(path, os.readlink(os.path.join(path, 'bootstrap.link')))
+		self._root_yml = load_bootstrap_yaml(root_path)
+
+		try:
+			with open(os.path.join(path, 'bootstrap-site.yml'), 'r') as f:
+				self._site_yml = yaml.load(f, Loader=global_yaml_loader)
+		except FileNotFoundError:
+			pass
+
+		self._parse_yml(os.path.join(path, os.readlink(os.path.join(path, 'bootstrap.link'))),
+				self._root_yml)
+
+	def _parse_yml(self, current_path, current_yml,
+			filter_sources = None, filter_tools = None, filter_pkgs = None, filter_tasks = None):
 		if 'imports' in current_yml and isinstance(current_yml['imports'], list):
 			if current_yml is not self._root_yml:
 				raise RuntimeError("Nested imports are not supported")
@@ -134,9 +203,10 @@ class Config:
 					raise RuntimeError("Unexpected data in import")
 
 				if 'from' in import_def:
+					import_path = os.path.join(os.path.dirname(current_path),
+							str(import_def['from']))
 					filter = dict()
-					with open(import_def['from'], 'r') as f:
-						import_yml = yaml.load(f, Loader=yaml.SafeLoader)
+					import_yml = load_bootstrap_yaml(import_path)
 					for f in ['sources', 'tools', 'packages', 'tasks']:
 						if "all_"+f in import_def:
 							filter[f] = None
@@ -144,12 +214,14 @@ class Config:
 							filter[f] = import_def[f]
 						else:
 							filter[f] = []
-					self._parse_yml(import_yml, filter_sources=filter['sources'], filter_tools=filter['tools'], filter_pkgs=filter['packages'], filter_tasks=filter['tasks'])
+					self._parse_yml(import_path, import_yml,
+							filter_sources=filter['sources'], filter_tools=filter['tools'],
+							filter_pkgs=filter['packages'], filter_tasks=filter['tasks'])
 				elif 'file' in import_def:
-					with open(import_def['file'], 'r') as f:
-						import_yml = yaml.load(f, Loader=yaml.SafeLoader)
-					self._parse_yml(import_yml)
-
+					import_path = os.path.join(os.path.dirname(current_path),
+							str(import_def['file']))
+					import_yml = load_bootstrap_yaml(import_path)
+					self._parse_yml(import_path, import_yml)
 
 		if 'sources' in current_yml and isinstance(current_yml['sources'], list):
 			for src_yml in current_yml['sources']:
@@ -519,6 +591,9 @@ class Source(RequirementsMixin):
 				args.append(self._this_yml['branch'])
 			if subprocess.call(args, cwd=self.source_dir, stdout=subprocess.DEVNULL) != 0:
 				return ItemState(missing=True)
+		elif 'svn' in self._this_yml:
+			if not os.path.isdir(self.source_dir):
+				return ItemState(missing=True)
 		else:
 			assert 'url' in self._this_yml
 			if not os.access(self.source_archive_file, os.F_OK):
@@ -984,18 +1059,7 @@ class RunTask(RequirementsMixin):
 		return 'task'
 
 def config_for_dir():
-	root_yml = None
-	with open('bootstrap.link', 'r') as f:
-		root_yml = yaml.load(f, Loader=yaml.SafeLoader)
-
-	site_yml = dict()
-	try:
-		with open('bootstrap-site.yml', 'r') as f:
-			site_yml = yaml.load(f, Loader=yaml.SafeLoader)
-	except FileNotFoundError:
-		pass
-
-	return Config(root_yml, site_yml)
+	return Config('')
 
 class EnvironmentComposer:
 	def __init__(self, cfg):
@@ -1182,6 +1246,11 @@ def fetch_src(cfg, src):
 		try_mkdir(src.source_dir)
 		args = ['hg', 'clone', source['hg'], src.source_dir]
 		subprocess.check_call(args)
+	elif 'svn' in source:
+		try_mkdir(src.source_dir)
+		args = ['svn', 'co', source['svn'], src.source_dir]
+		print (args)
+		subprocess.check_call(args)
 	else:
 		assert 'url' in source
 
@@ -1226,6 +1295,12 @@ def checkout_src(cfg, src, settings):
 		else:
 			args.append(source['branch'])
 		subprocess.check_call(args, cwd=src.source_dir)
+	elif 'svn' in source:
+		args = ['svn', 'update']
+		if 'rev' in source:
+			args.append('-r')
+			args.append(source['rev'])
+		subprocess.check_call(args, cwd=src.source_dir)
 	else:
 		if src.source_archive_format == 'raw' and 'filename' in source:
 			shutil.copyfile(src.source_archive_file, os.path.join(src.sub_dir, src.name, source['filename']))
